// Copyright 1997, John Webster Small
// All Rights Reserved

#ifndef ptr_h
#define ptr_h

// To port: typedef Monitor and Lock
#include <classlib\thread.h>
typedef TCriticalSection Monitor;
typedef TCriticalSection::TLock Lock;
/*
class Monitor { public: Monitor() {} };
class Lock { public: Lock(Monitor& m) {} };
*/

template <class TYPE>
class Linkable
{
    unsigned links;
    TYPE * _T;
  public:
    Monitor m;
    Linkable(TYPE * T) : links(1), _T(T), m() {}
    virtual ~Linkable() {}
    operator TYPE * () { return _T; }
    void link()
    {
      Lock lock(m);
      if (links + 1 == 0)
        throw "overlinked";
      ++links;
    }
    unsigned unlink()
    {
      Lock lock(m);
      return --links;
    }
    static void assign
      (Linkable<TYPE > *& L, Linkable<TYPE > * L_)
    {
      if (L != L_)  {
        if (L_)
          L_->link();
        if (L && !L->unlink())
          delete L;
        L = L_;
      }
    }
};

template <class TYPE>
class Ptr
{
  protected:
    Monitor m;
    Linkable<TYPE > * L;
    void assign(Ptr<TYPE >& p)
    {
      Lock lock(m);
      Linkable<TYPE >::assign(L,p.L);
    }
  private:
    class Instance : public Linkable<TYPE >
    {
        TYPE T;
      public:
        Instance() : Linkable<TYPE >(&T), T() {}
    };
  public:
    Ptr() : m(), L(0) { L = new Instance(); }
    Ptr(Ptr<TYPE >& p) : m(), L(0) { assign(p); }
    Ptr(Linkable<TYPE > * L_) : m(), L(L_) {}
    ~Ptr() { Linkable<TYPE >::assign(L,0); }
    Ptr<TYPE >& operator=(Ptr<TYPE >& p)
      { assign(p); return *this; }
    Monitor& getMonitor() { return L->m; }
    TYPE * operator->()
    {
      Lock lock(m);
      return *L;
    }
};

template <class TYPE, class ROOT>
class DPtr : public Ptr<ROOT >
{
    class Instance : public Linkable<ROOT >
    {
        TYPE T;
      public:
        Instance() : Linkable<ROOT >(&T), T() {}
    };
  public:
    DPtr() : Ptr<ROOT >(new Instance()) {}
    DPtr(Ptr<ROOT >& p) : Ptr<ROOT >(p) {}
    DPtr<TYPE,ROOT >& operator=(Ptr<ROOT >& p)
      { assign(p); return *this; }
    TYPE * operator->()
    {
      Lock lock(m);
      return dynamic_cast<TYPE *>((ROOT*)*L);
    }
};

#endif
